import * as mockttp from 'mockttp';
import {
  CompletedRequest,
  CompletedResponse,
  CompletedBody,
  TimingEvents,
} from 'mockttp';
// import { getCA, CAOptions } from 'mockttp/dist/util/tls';
import { Subject } from '../../../lib/async-tools';
import { AbortSignal, AbortController } from 'node-abort-controller'; // remove when Node v14 is out of LTS
import { pki, md } from 'node-forge';
import { randomBytes } from 'crypto';
import { Readable } from 'stream';
import EventEmitter from 'events';
import http from 'http';
import http2 from 'http2';
import tls from 'tls';
import net from 'net';
import { URL } from 'url';
import * as httpolyglot from '@httptoolkit/httpolyglot';
import { Server as ConnectServer } from 'connect';
import portfinder from 'portfinder';
import globalLog from 'log';
import invariant from 'ts-invariant';
import chalk from 'chalk';
type Logger = typeof globalLog;

export const log: Logger = globalLog.get('captures:streams:sources:proxy'); // export so it can be enabled in testing

export interface ProxyInteractions
  extends AsyncIterable<ProxySource.Interaction> {}

export class ProxyInteractions {
  static async create(
    targetHost: string,
    abort: AbortSignal, // required, we don't want to ever let a proxy run indefinitely
    options: {
      ca?: ProxyCertAuthority;
      targetCA?: Array<{ cert: Buffer | string }>;
    } = {}
  ): Promise<[ProxyInteractions, string, string]> {
    let { host, protocol } = new URL(targetHost);
    if (targetHost.includes('/')) {
      // accept urls to be passed in rather than pure hosts
      targetHost = host;
    }

    if (protocol)
      invariant(
        targetHost.match(/^([a-zA-Z0-9\-]+\.)*[a-zA-Z0-9\-]+(:\d+)?$/),
        'targetHost must be a valid host (hostname:port)'
      ); // port optional

    if (protocol === 'https:') {
      console.log(
        chalk.gray(
          'Target uses TLS. If traffic is not being captured run "oas setup-tls"'
        )
      );
    }

    const capturingProxy = mockttp.getLocal({
      cors: false,
      debug: false,
      recordTraffic: false,
      https: options.ca && {
        cert: options.ca.cert.toString(),
        key: options.ca.key.toString(),
      },
    });

    let forwardedHosts = [targetHost];
    await capturingProxy
      .forAnyRequest()
      .always()
      .matching((request: CompletedRequest) => {
        // our own matching, adapted from mockttp's HostMatcher, so we can forward
        // direct requests to the proxy to the target as well
        const parsedUrl = new URL(request.url);

        let result = false;
        for (let host of forwardedHosts) {
          if (
            (host.endsWith(':80') && request.protocol === 'http') ||
            (host.endsWith(':443') && request.protocol === 'https')
          ) {
            // On default ports, our URL normalization erases an explicit port, so that a
            // :80 here will never match anything. This handles that case: if you send HTTP
            // traffic on port 80 then the port is blank, but it should match for 'hostname:80'.
            result =
              result ||
              (parsedUrl.hostname === host.split(':')[0] &&
                parsedUrl.port === '');
          } else {
            result = result || parsedUrl.host === host;
          }
          if (result) break;
        }

        return result;
      })
      .thenPassThrough({
        beforeRequest: onTargetedRequest,
        forwarding: {
          targetHost,
          updateHostHeader: true,
        },
        trustAdditionalCAs: options.targetCA || [],
      });

    await capturingProxy.forUnmatchedRequest().thenPassThrough({
      beforeRequest(capturedRequest) {
        log.info('proxying request to ' + capturedRequest.url);
      },
    });
    await capturingProxy.forAnyWebSocket().thenPassThrough();

    const interactions = new Subject<ProxySource.Interaction>();

    const requestsById = new Map<string, ProxySource.Request>();
    function onTargetedRequest(capturedRequest: CompletedRequest) {
      const {
        matchedRuleId,
        remoteIpAddress,
        remotePort,
        tags,
        body,
        timingEvents,
        ...rest
      } = capturedRequest;

      const request = {
        ...rest,
        body: { buffer: body.buffer },
        timingEvents: timingEvents as TimingEvents,
      };

      requestsById.set(request.id, request);
    }

    function onResponse(capturedResponse: CompletedResponse) {
      const { id } = capturedResponse;
      const request = requestsById.get(id);
      if (!request) return;

      const { tags, body, timingEvents, ...rest } = capturedResponse;

      const response = {
        ...rest,
        body: { buffer: body.buffer },
        timingEvents: timingEvents as TimingEvents,
      };
      interactions.onNext({
        request,
        response,
      });

      requestsById.delete(id);
    }

    await capturingProxy.on('response', onResponse);

    abort.addEventListener('abort', onAbort);

    interactions.finally(() => {
      abort.removeEventListener('abort', onAbort);
    });

    function onAbort(e) {
      interactions.onCompleted();
    }

    let transparentPort = await portfinder.getPortPromise({
      port: 8000,
      stopPort: 8999,
    });
    await capturingProxy.start({
      startPort: transparentPort + 1,
      endPort: transparentPort + 999,
    });
    forwardedHosts.push(`localhost:${capturingProxy.port}`);
    log.info('capturing proxy started at %s', capturingProxy.url);

    // sits in front of the capturing proxy to only direct target host traffic
    // and transparently forward the rest
    // @ts-ignore
    let capturingApp = capturingProxy.app as ConnectServer;
    // @ts-ignore
    let capturingServer = capturingProxy.server as net.Server;
    // @ts-ignore
    const tlsServer = capturingServer._tlsServer as tls.Server &
      tls.SecureContextOptions;

    const transparentProxy = new TransparentProxy(
      targetHost,
      capturingApp,
      capturingServer,
      {
        https: options.ca && {
          ca: [options.ca.cert],
          cert: tlsServer.cert,
          key: tlsServer.key,
        },
      }
    );

    await transparentProxy.start(transparentPort);
    forwardedHosts.push(`localhost:${transparentProxy.port}`);
    log.info('transparent proxy started at %s', transparentProxy.url);

    const stream = (async function* () {
      yield* interactions.iterator;
      await capturingProxy.stop();
      await transparentProxy.stop();
    })();

    return [stream, transparentProxy.url!, capturingProxy.url];
  }
}

export declare namespace ProxySource {
  interface Interaction {
    request: Request;
    response: Response;
  }

  interface Request
    extends Omit<
      CompletedRequest,
      'matchedRuleId' | 'remoteIpAddress' | 'remotePort' | 'tags' | 'body'
    > {
    timingEvents: TimingEvents;
    body: Body;
  }
  interface Response extends Omit<CompletedResponse, 'tags' | 'body'> {
    timingEvents: TimingEvents;
    body: Body;
  }

  type Body = Pick<CompletedBody, 'buffer'>;
}

export interface ProxyCertAuthority {
  cert: string;
  key: string;
  keyLength?: number;
}

export class ProxyCertAuthority {
  static async generate(): Promise<ProxyCertAuthority> {
    const keyPair = await new Promise<pki.rsa.KeyPair>((resolve, reject) => {
      pki.rsa.generateKeyPair({ bits: 2048 }, (err, keypair) => {
        if (err) {
          reject(err);
        } else {
          resolve(keypair);
        }
      });
    });

    const cert = pki.createCertificate();
    cert.publicKey = keyPair.publicKey;
    cert.serialNumber = generateSerialNumber();

    cert.validity.notBefore = new Date();
    cert.validity.notBefore.setDate(cert.validity.notBefore.getDate() - 1); // account for wonky time keeping
    cert.validity.notAfter = new Date();
    cert.validity.notAfter.setDate(cert.validity.notAfter.getDate() + 30);

    cert.setSubject([
      {
        name: 'commonName',
        value: `Optic CLI CA (locally generated ${new Date().toISOString()})`,
      },
      { name: 'countryName', value: 'US' },
      { name: 'stateOrProvinceName', value: 'NY' },
      { name: 'localityName', value: 'New York City' },
      { name: 'organizationName', value: 'Optic Labs Corporation' },
      { name: 'organizationalUnitName', value: 'https://useoptic.com' },
    ]);
    cert.setIssuer(cert.subject.attributes); // self-signed

    cert.setExtensions([
      { name: 'basicConstraints', cA: true, critical: true },
      {
        name: 'keyUsage',
        keyCertSign: true,
        digitalSignature: true,
        nonRepudiation: true,
        cRLSign: true,
        critical: true,
      },
      {
        name: 'extKeyUsage',
        serverAuth: true,
        clientAuth: true,
      },
      {
        name: 'subjectKeyIdentifier',
      },
    ]);

    cert.sign(keyPair.privateKey, md.sha256.create());

    return {
      cert: pki.certificateToPem(cert),
      key: pki.privateKeyToPem(keyPair.privateKey),
    };
  }

  static hasExpired(self: ProxyCertAuthority, dateTime: Date): boolean {
    const cert = pki.certificateFromPem(self.cert);

    return (
      dateTime < cert.validity.notBefore || dateTime > cert.validity.notAfter
    );
  }

  static readableCert(self: ProxyCertAuthority): Readable {
    return Readable.from(Buffer.from(self.cert));
  }
}

function generateSerialNumber(): string {
  // hexadecimal serial number of at most 20 octets, and preferably positive.
  // starting with A should get a positive number
  return 'A' + randomBytes(18).toString('hex').toUpperCase();
}

class TransparentProxy {
  server: httpolyglot.Server;
  port?: number;
  url?: string;
  tunnelAbort: AbortController;
  tlsEnabled: boolean;
  serverDestroy: (cb: (err?: Error) => void) => void;

  constructor(
    targetHost: string,
    captureRequest: http.RequestListener,
    capturingServer: net.Server,
    options: {
      https?: tls.SecureContextOptions;
    } = {}
  ) {
    this.server = httpolyglot.createServer(
      {
        ALPNProtocols: ['http/1.1'],
        ...(options.https ? options.https : {}),
      },
      captureRequest
    );

    this.serverDestroy = destroyCommandForServer(this.server);

    this.tlsEnabled = !!options.https;

    const tunnelAbort = (this.tunnelAbort = new AbortController()); // control aborting of tunnels separately
    // small hack to prevent warnings, because the AbortController polyfill doesn't setup EventEmitter according to spec
    // @ts-ignore
    if (tunnelAbort.signal.eventEmitter) {
      // @ts-ignore
      (tunnelAbort.signal.eventEmitter as EventEmitter).setMaxListeners(
        Infinity
      );
    }

    let [targetHostname, targetPort] = targetHost.split(':');

    this.server.on(
      'connect',
      (
        req: http.IncomingMessage | http2.Http2ServerRequest,
        resOrSocket: net.Socket | http2.Http2ServerResponse,
        reqHead: Buffer
      ) => {
        if (resOrSocket instanceof net.Socket) {
          let clientSocket = resOrSocket;
          // clients disconnecting causes errors, but there's not much to do for us
          // than the default behaviour of stopping the tunnel
          clientSocket.once('error', (err) => {
            if (isTransitiveSocketError(err)) {
              // connections get broken sometimes, nothing special
              clientSocket.destroy();
              log.info(
                'Handled error on client socket by destroying it: %s',
                err
              );
              return;
            }

            this.server.emit('error', err); // forward any errors we can't handle
          });
          destroySocketOnAbort(clientSocket, tunnelAbort.signal);

          if (!req.url) {
            // TODO: see if we could go just of the `host` header
            clientSocket.write(
              `HTTP/${req.httpVersion} 400 Bad Request\r\n\r\n`,
              'utf-8'
            );
            clientSocket.end();
            return;
          }

          const { port, host, hostname } = new URL(`http://${req.url}`);
          if (
            host !== targetHost &&
            !(
              targetHostname === hostname &&
              (port === targetPort ||
                (!targetPort && port === '80') ||
                (!targetPort && port === '443'))
            )
          ) {
            // transparently tunnel all non-target traffic directly through TCP sockets
            log.info('transparently tunneling connection to %s', host);
            const serverSocket = net.connect(
              (port && parseInt(port)) || 80,
              hostname,
              () => {
                log.info('tunnel to %s established', host);
                clientSocket.once('error', (err) => {
                  serverSocket.destroy();
                });

                clientSocket.write(
                  `HTTP/${req.httpVersion} 200 Connection Establishe d\r\nProxy-agent: Optic Transparent proxy\r\n\r\n`
                );
                serverSocket.write(reqHead);
                clientSocket.pipe(serverSocket);
                serverSocket.pipe(clientSocket);
              }
            );

            serverSocket.once('error', (err) => {
              let errorCode = isCodedError(err) ? err.code : null;

              if (
                (errorCode === 'ENOTFOUND' || errorCode === 'ECONNREFUSED') &&
                serverSocket.connecting
              ) {
                clientSocket.write(
                  `HTTP/${req.httpVersion} 502 Bad Gateway\r\nProxy-agent: Optic Transparent proxy\r\n\r\n`
                );
                log.notice(
                  'tunnel to %s could not be established: %s',
                  host,
                  errorCode
                );
                return;
              } else if (errorCode === 'ETIMEDOUT' && serverSocket.connecting) {
                clientSocket.write(
                  `HTTP/${req.httpVersion} 504 Gateway Timeout\r\nProxy-agent: Optic Transparent proxy\r\n\r\n`
                );
                log.notice(
                  'tunnel to %s could not be established: %s',
                  host,
                  errorCode
                );
                return;
              } else if (isTransitiveSocketError(err)) {
                // connections get broken sometimes, nothing special
                clientSocket.destroy();
                log.info('tunnel to %s interrupted: %s', host, errorCode);
                return;
              }

              this.server.emit('error', err); // forward error to transparentProxy server, give downstream a chance to handle it
            });
            destroySocketOnAbort(serverSocket, tunnelAbort.signal);
          } else {
            log.notice('capturing connection to target (%s)', host);
            // tunnel target traffic to the capturing proxy
            // @ts-ignore
            clientSocket.write(
              `HTTP/${req.httpVersion} 200 Connection Established\r\nProxy-agent: Optic Transparent proxy\r\n\r\n`
            );
            capturingServer.emit('connection', clientSocket);
          }
        } else {
          throw new Error('HTTP/2 CONNECT unimplemented');
        }
      }
    );
  }

  start(port: number): Promise<void> {
    const server = this.server;

    return new Promise<void>((resolve, reject) => {
      const onListening = () => {
        cleanup();
        this.port = port;
        this.url = `${this.tlsEnabled ? 'https' : 'http'}://localhost:${port}`;
        resolve();
      };

      const onListenError = (err: Error) => {
        let errorCode = isCodedError(err) && err.code;
        if (errorCode && errorCode === 'EADDRINUSE') {
          // port got taken since we last found it as un-used, try again
          log.notice('attempting another port for transparentProxy');
          portfinder
            .getPortPromise({
              port: 8000,
              stopPort: 8999,
            })
            .then((availablePort) => {
              port = availablePort;
              server.close();
              server.listen(port);
            });
        } else {
          cleanup();
          reject(err);
        }
      };

      const cleanup = () => {
        this.server.removeListener('listening', onListening);
        this.server.removeListener('error', onListenError);
      };

      server.on('listening', onListening);
      server.on('error', onListenError);
      server.listen(port);
    });
  }

  stop(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      this.serverDestroy((err) => {
        // stops accepting new connections, waits for tunnels to finish
        if (err) {
          reject(err);
        } else {
          resolve();
        }
      });
      this.tunnelAbort.abort(); // abort all open tunnels so tunneling proxy can close
    });
  }
}

function destroySocketOnAbort(socket: net.Socket, abort: AbortSignal) {
  socket.once('close', (err) => {
    abort.removeEventListener('abort', onAbort);
  });

  abort.addEventListener('abort', onAbort);

  function onAbort() {
    if (!socket.destroyed) socket.destroy();
  }
}

function isCodedError(
  error: Error & { code?: string }
): error is Error & { code: string } {
  return typeof error.code === 'string';
}

function isTransitiveSocketError(err: Error) {
  return isCodedError(err) && transitiveSocketErrors.includes(err.code);
}

const transitiveSocketErrors = Object.freeze([
  'ECONNRESET',
  'ENOTFOUND',
  'EPIPE',
  'ETIMEDOUT',
]);

function destroyCommandForServer(server: httpolyglot.Server) {
  let connections = {};

  server.on('connection', function (conn) {
    var key = conn.remoteAddress + ':' + conn.remotePort;
    connections[key] = conn;
    conn.on('close', function () {
      delete connections[key];
    });
  });

  return function (cb: (err?: Error) => void) {
    server.close(cb);
    for (var key in connections) connections[key].destroy();
  };
}
